// java

package edu.washington.cac.calendar.data;

import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Calendar;
import java.util.Iterator;

import edu.washington.cac.calendar.CalLogFactory;
import edu.washington.cac.calendar.CalProperties;
import edu.washington.cac.calendar.MyCalendar;
import edu.washington.cac.calendar.db.Caldata;
import edu.washington.cac.calendar.filter.PublicCalendars;

/**
  A global singleton set of all public events.  This is a thin wrapper 
  around an Events object with the appropriate methods synchronized,
  and other methods disallowed.

  @author Greg Barnes
  @version 1.0
 */
public class PublicEvents extends Events
{
  /** The set of public calendars */
  private static PublicCalendars pc = null;

  /** The single instance of this class */
  private static PublicEvents pe = null;

  /** Initialize the singleton */
  static {
    initializePublicEvents();
  }

  /**
    Initialize the singleton set of public events
   */
  private static synchronized void initializePublicEvents()
  {
    if (pe == null || pc == null) {
      try {
        pe = new PublicEvents();
        pc = new PublicCalendars(new Caldata());
        pc.loadCalendars();
      } catch (Exception e) {
        e.printStackTrace(new CalLogFactory().create().getPrintStream());
        new CalLogFactory().create().println("Could not create public events");
        pe = null;
        pc = null;
      }
    }
  }

  /** 
    Get the set of public events
    @return the set of public events
    */
  public static PublicEvents getPublicEvents()
  {
    if (pe == null) {
      initializePublicEvents();

      if (pe == null) {
        throw new RuntimeException("Could not get public events");
      }
    }

    return pe;
  }

  /** Get the public calendars 
    @return The public calendars
   */
  public static PublicCalendars getPublicCalendars()
  {
    if (pc == null) {
      initializePublicEvents();

      if (pc == null) {
        throw new RuntimeException("Could not get public calendars");
      }
    }

    return pc;
  }

  /** Get the public keywords 
    @return The public keywords
   */
  public static Keywords getPublicKeywords()
  {
    return pe.getKeywords();
  }

  /** Get the public sponsors 
    @return The public sponsors
   */
  public static Sponsors getPublicSponsors()
  {
    return pe.getSponsors();
  }

  /** Get the public locations 
    @return The public locations
   */
  public static Locations getPublicLocations()
  {
    return pe.getLocations();
  }

  /** Initialize the set, including loading locations, sponsors and keywords
    @exception SQLException if the auxiliary items cannot be loaded
    @exception CaldataException if the auxiliary items contain bad data
   */
  private PublicEvents() throws SQLException, CaldataException
  {
    super(new User(), new Caldata(), true);   // loads locations
    setKeywords(new Keywords(new User(), new Caldata(), true)); // load keywords
    setSponsors(new Sponsors(new User(), new Caldata(), true)); // load sponsors

    if (CalProperties.defaultIsGuest()) {
//      loadAllEvents();
    }
  }

  /* note that adds are allowed to this set, to populate the set. 
     Other manipulations are not.
   */
  /** not allowed 
    @exception IllegalStateException Always thrown
   */
  public void addEventRef(EventRef er, MyCalendar mc) 
     throws IllegalStateException
  {
    throw new IllegalStateException("cannot add event refs");
  }

  /** not allowed 
    @exception IllegalStateException Always thrown
   */
  public void delete(Cacheable c) throws IllegalStateException
  {
    throw new IllegalStateException("deletes not allowed"); 
  }

  /** not allowed 
    @exception IllegalStateException Always thrown
   */
  public void replace(Cacheable c) throws IllegalStateException
  {
    throw new IllegalStateException("updates not allowed"); 
  }

  /* All other public methods are synchronized to avoid inconsistent views
     of the data (note:  public methods that are simply a call to
     another public method with different parameters are left unsynchronized
   */
  /**
    @return The number of items in the Cache
   */
  public synchronized int size() 
  {
    return super.size();
  }

  /**
    @return All the items in the cache
   */
  public synchronized Iterator elements() 
  {
    return super.elements();
  }

  /**
    Add an item to the cache
    @param c Item to add
    @param saveToDB Save to database as well?
    @exception SQLException Should not occur
   */
  public synchronized void add(Cacheable c, boolean saveToDB) 
      throws SQLException
  {
    super.add(c, saveToDB);
  }

  /**
    get an item from the Cache 

    @param id ID of the item
    @return The Cacheable item with the same id
    @exception NoSuchItemException if no item exists with that id
   */
  public synchronized Cacheable get(int id) throws NoSuchItemException
  {
    return super.get(id);
  }

  /**
    @return If the key refers to an object in the cache
    @param o key to check
   */
  public synchronized boolean containsKey(Object o) 
  {
    return super.containsKey(o);
  }

  /**
    @return all the elements in the cache, in sorted order
   */
  public synchronized Iterator sortedElements() 
  {
    return super.sortedElements();
  }

  /** Set the keywords cache 
    @param ks new Keywords cache
   */
  public synchronized void setKeywords(Keywords ks)
  {
    super.setKeywords(ks);
  }

  /** Set the Locations cache 
    @param ls new Locations cache
   */
  public synchronized void setLocations(Locations ls)
  {
    super.setLocations(ls);
  }

  /** Set the sponsors cache 
    @param ss new Sponsors cache
   */
  public synchronized void setSponsors(Sponsors ss)
  {
    super.setSponsors(ss);
  }

  /** 
      Get the set of associated locations for a user
      @return the set of associated locations for a user
    */
  public synchronized Locations getLocations() 
  {
    return super.getLocations();
  }

  /** 
    Get the set of associated sponsors for a user
    @return the set of associated sponsors for a user
    */
  public synchronized Sponsors getSponsors() 
  {
    return super.getSponsors();
  }

  /** 
    Get the set of associated keywords for a user
    @return the set of associated keywords for a user
    */
  public synchronized Keywords getKeywords() 
  {
    return super.getKeywords();
  }

  /** 
    Get the set of associated keywords for a user
    @return the set of associated keywords for a user
    @exception SQLException if there is a problem loading from the database
    @exception CaldataException if the database contains bad data
    */
  public synchronized void loadAllEvents() throws SQLException, CaldataException
  {
    super.loadAllEvents();
  }

  /** 
    Get the events for a user on a given day, sorted in an appropriate
    way

    @param mc Date, represented as a <code>MyCalendar</code>
    @param loadKeywords should the keywords be loaded (if the events are?)

    @return the events for a user on a given day, sorted
    @exception SQLException if there is a problem loading from the database
    @exception CaldataException if the database contains bad data
  */
  public synchronized Event[] oneDaysEvents(MyCalendar mc, boolean loadKeywords)
      throws SQLException, CaldataException
  {
    return super.oneDaysEvents(mc, loadKeywords);
  }

  /** 
    Alert the object that events from many days are about to be 
    requested.  This need not be called, but should speed up a
    series of calls to <code>oneDaysEvents()</code> if it is

    @param start The first date in a series of dates
    @param end The last date in a series of dates
    @exception SQLException if we have trouble loading data
    @exception CaldataException if the data we try to load is bad
  */
  public synchronized void prepareManyDays(Calendar start, 
                                           Calendar end) 
      throws SQLException, CaldataException
  {
    super.prepareManyDays(start, end);
//new edu.washington.cac.myuw.servlet.Memory().printStatus(new CalLogFactory().create());
//new CalLogFactory().create().println(size());
  }

  /** load a result set containing events for one day
      @param mc The day in question
      @param rs The <code>java.sql.ResultSet</code> containing the events
      @exception SQLException if there is database trouble
      @exception CaldataException if the data we get back is bad
    */
  public synchronized void load(MyCalendar mc, ResultSet rs) 
      throws SQLException, CaldataException
  {
    super.load(mc, rs);
  }

  /**
    Was this event loaded by this cache?
    @param e Event
    @return Was this event loaded by this cache?
   */
  protected boolean loadedByThisCache(Event e)
  {
    return e.isPublic();
  }

  /**
    Do an initial public events load
    @param args Ignored
   */
  public static void main(String[] args)
  {
    getPublicEvents();
  }
}
