// java

package edu.washington.cac.calendar.db;

import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;

import edu.washington.cac.calendar.CalLogFactory;
import edu.washington.cac.gsbarnes.util.Timer;
import edu.washington.cac.myuw.servlet.Log;
import edu.washington.cac.myuw.servlet.CanUseLog;

/** A database connection, statements and result sets 

    This object is not thread safe; different threads should use different
    copies of this object.

    The close method should not be called on the objects returned by this
    object.  Instead, this object's close method should be called when
    a statement has been completely processed.

    @author Greg Barnes
    @version 2.0
  */
public class CalConnection {
  static {    // initialize the driver
    try {
      Class.forName(DBResources.driverClassName()).newInstance();
    } catch (Exception e) {
      new CalLogFactory().create().println("Unable to load JDBC driver " + 
                                           DBResources.driverClassName());
      e.printStackTrace(new CalLogFactory().create().getPrintStream());
    }
  }

  /** Name of the database to use.  Should be either 
    <code>DBResources.PROD</code> or <code>DBResources.DEV</code>
   */
  private String dbName;

  /** The connection we are currently using */
  private Connection conn = null;

  /** The statement we are currently using */
  private Statement stmt = null;

  /** The results of the latest query */
  private ResultSet rs = null;

  /** A file to write message to */
  private Log log;

  /** Create a new connection
    @param log File for log messages
   */
  public CalConnection(Log log) 
  {
    this(log, DBResources.defaultDb());
  }

  /** Create a new connection
    @param log File for log messages
    @param dbName Name of the database to use.  Should be either 
      <code>DBResources.PROD</code> or <code>DBResources.DEV</code>
   */
  public CalConnection(Log log, String dbName) 
  {
    this.log = log;
    this.dbName = dbName;
  }

  /** Ensure the connection is open
    @exception SQLException If we have to open a connection, and there's
      a problem
   */
  private void setConnection() 
    throws SQLException
  {
    if (conn == null) {
      if (DBResources.user() == null) {   // connect with just URL
        conn = DriverManager.getConnection(DBResources.jdbcUrl(this.dbName));
      } else {				// connect with username and password
        conn = DriverManager.getConnection(DBResources.jdbcUrl(this.dbName), 
                                           DBResources.user(), 
                                           DBResources.password());
      }
      
      if (conn instanceof edu.washington.cac.myuw.servlet.CanUseLog) {
        ((CanUseLog) conn).setLog(log);
      }
    }
  }

  /** Get a new <code>PreparedStatement</code>
    @param sql SQL for the prepared statemnt
    @exception SQLException if there's a problem
   */
  public PreparedStatement getPreparedStatement(String sql) throws SQLException
  {
    closeStatement();
    setConnection();
    stmt = conn.prepareStatement(sql);
    return (PreparedStatement) stmt;
  }
      
  /** Create a new statement
    @exception SQLException if there's a problem
   */
  private void setStatement() throws SQLException
  {
    closeStatement();
    setConnection();
    stmt = conn.createStatement();
  }

  /** Execute a query
    @param sql Query to execute
    @return The result set of the query
    @exception SQLException if there's a problem
   */
  public ResultSet executeQuery(String sql) throws SQLException
  {
    closeResultSet();
    setStatement();
//Timer timer = new Timer();
//timer.start();
    rs = stmt.executeQuery(sql);
//timer.printElapsedTime(log, "executeQuery " + sql + ":");

    return rs;
  }

  /** Close the current statement */
  private void closeStatement() 
  {
    try {
      if (stmt != null) {
        stmt.close();
      }
    } catch (SQLException e) {      // don't care
    } finally {
      stmt = null;
    }
  }

  /** Close the current connection */
  private void closeConnection() 
  {
    try {
      if (conn != null) {
        conn.close();
      }
    } catch (SQLException e) {      // don't care
    } finally {
      conn = null;
    }
  }

  /** Close the current result set */
  private void closeResultSet()
  {
    try {
      if (rs != null) {
        rs.close();
      }
    } catch (SQLException e) {      // don't care
    } finally {
      rs = null;
    }
  }

  /** Close all db-related things that may be open (connections, statements,
     and result sets
   */
  public void close() {
    closeResultSet();
    closeStatement();
    closeConnection();
  }
}
