package edu.washington.cac.myuw.servlet;

import java.io.*;
import java.util.*;

import javax.servlet.*;
import javax.servlet.http.*;

/** A basic servlet with a log and some useful routines 
 @author FangLin, Greg Barnes
 @version 2.0
*/
public class UWServlet extends HttpServlet {
  /** default printable id of the servlet */
  private static final String DEFAULT_ID = "Unnamed servlet";

  /** printable id of the servlet */
  protected String id = DEFAULT_ID;

  /** a factory to create logs */
  private LogFactory logFactory = null;

  /** Create a new servlet 
    @param logFactory A factory to create logs
   */
  protected UWServlet(LogFactory logFactory) {
    this.logFactory = logFactory;
  }

  /** Create a new servlet with the default log factory
   */
  protected UWServlet() {
    this.logFactory = new LogFactory();
  }

  /** @return The servlet's log factory */
  protected LogFactory getLogFactory() {
    return this.logFactory;
  }

  /** @return A default log for this servlet */
  protected Log createLog() {
    return logFactory.create();
  }

  /** @return the value of the HTTP header Referer */
  protected String getReferer (HttpServletRequest req) {
    return req.getHeader("Referer");
  }

    private static final String HTTP = "http";
    private static final String HTTPS = "https";
    private static final String PROTOCOL_END = "://";
    private static final String ROOTED_START = "/";

    protected static boolean isRootedRelativeUrl(String url) {
	return url.startsWith(ROOTED_START);
    }

    protected static boolean isRelativeUrl(String url) {
	return url.indexOf(PROTOCOL_END) == -1;
    }

    protected static boolean isHttps (HttpServletRequest req) {
        return req.getScheme().equals(HTTPS);
    }

  private static final int DEF_SSL_PORT = 443;
  private static final String DEF_SSL_PORTS = "443";
  private static final int DEF_SSL_PORTL = 3;

  // return protocol, server name, port number, and servlet path, and servlet name
  public static StringBuffer getReqUrl (HttpServletRequest req, 	
					boolean forceHttps)
    {
      if (req == null) {
	throw new IllegalArgumentException("UWServlet::getReqURL: req is null.");
      }

      StringBuffer requrl = HttpUtils.getRequestURL(req);
      if (forceHttps && !isHttps(req)) {
	requrl.insert(4, 's');
      }

      //remove the port number if it is the default SSL port
      int port = req.getServerPort();
      //System.err.println(port);
      String urlString = requrl.toString();
      int startIndex=0, endIndex=0;
      if (port == DEF_SSL_PORT) {
        startIndex = urlString.indexOf(DEF_SSL_PORTS);
	endIndex =  startIndex + DEF_SSL_PORTL;
	if (startIndex != -1) {
	  //System.err.println(startIndex);
	  //System.err.println(endIndex);
	  requrl.delete(startIndex-1, endIndex);
	}   	
      }
      
      return requrl;
    }

  // return protocol, server name, port number, and servlet path
  protected static StringBuffer getReqUrlPath (HttpServletRequest req, 	
					       boolean forceHttps)
    {
      if (req == null) {
	throw new IllegalArgumentException("UWServlet::getReqURL: req is null.");
      }
      return trimServlet(getReqUrl(req, forceHttps));
    }
  
  protected static StringBuffer trimServlet (StringBuffer requrl) {
    // chop off servlet name
    requrl.setLength(requrl.toString().lastIndexOf("/")+1);
    return requrl;
  }

  // return protocol, server name, port number
  protected static StringBuffer getReqUrlNoPath (HttpServletRequest req,
						 boolean forceHttps)
    {
      if (req == null) {
	throw new IllegalArgumentException("UWServlet::getReqURLNoPath: req is null.");
      }
      StringBuffer requrl = getReqUrl(req, forceHttps);
      String urlString = requrl.toString();
      requrl.setLength(urlString.indexOf(ROOTED_START,8)+1);
      
      return requrl;
    }

  /*
    protected static StringBuffer protocol(HttpServletRequest req, 	
					   boolean forceHttps) 
    {
        StringBuffer ret = new StringBuffer();

        if ((req != null && isHttps(req)) || forceHttps) {
	    ret.append(HTTPS);
        } else {
	    ret.append(HTTP);
        }

        ret.append(PROTOCOL_END);
	return ret;
    }

      // get prefix for any absolute URL on this server
    protected static String absoluteUrlPrefix(HttpServletRequest req,
					      boolean forceHttps)
    {
        StringBuffer ret = protocol(req, forceHttps);
        ret.append(req.getServerName());
        return ret.toString();
    }

    protected static String currentPath(HttpServletRequest req,
					      boolean forceHttps)
    {
	return absoluteUrlPrefix(req, forceHttps) + req.getServletPath();
	//getServletPath() doesn't return the correct path. -fang
    }
  */
    protected static String redirectUrl(HttpServletRequest req, 
					String location, 
					boolean forceHttps)
    {
        if (isRootedRelativeUrl(location)) {
	    return getReqUrlNoPath(req, forceHttps).append(location).toString();

	} else if (isRelativeUrl(location)) {
	    return getReqUrlPath(req, forceHttps).append(location).toString();
        } else {
	    return location;
        }
    }
     
      /* A more programmer-friendly version of sendRedirect which
         a) calls encodeRedirectUrl if encode is true
         b) turns apparently relative URLs into absolute URLs
      */
    public static void sendRedirect(HttpServletRequest req, 
				    HttpServletResponse res,
				    String location,
				    boolean forceHttps,
				    boolean encode) 
      throws IOException
    {
	if (encode) {
            res.sendRedirect(
	      res.encodeRedirectUrl(redirectUrl(req, location, forceHttps)));
	} else {
            res.sendRedirect(redirectUrl(req, location, forceHttps));
 	}
    }

      /* A more programmer-friendly version of sendRedirect which
         a) remembers to call encodeRedirectUrl
         b) turns apparently relative URLs into absolute URLs
      */
    public static void sendRedirect(HttpServletRequest req, 
				    HttpServletResponse res,
				    String location,
				    boolean forceHttps) 
      throws IOException
    {
	sendRedirect(req, res, location, forceHttps, true);
    }

      /* returns the part of the servlet URL before the servletPath
       */
    public static String preServletPath(HttpServletRequest req) {
	return req.getRequestURI().substring(0, 
	  req.getRequestURI().lastIndexOf("/"));
    }


}











