package pubevents;

import java.io.IOException;
import java.io.PrintWriter;
import java.sql.ResultSet;
import java.sql.SQLException;


import javax.servlet.ServletConfig;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import edu.washington.cac.calendar.CalLogFactory;
import edu.washington.cac.calendar.db.Fields;
import edu.washington.cac.calendar.db.Tables;
import edu.washington.cac.myuw.servlet.ServletSupport;
import edu.washington.cac.myuw.servlet.UWServlet;
import edu.washington.cac.myuw.servlet.session.InvalidSession;

import pubevents.util.PEMessageMaker;
import pubevents.util.PrintWebPage;
import pubevents.addevents.SqlStatements;

/** The base class of most of the UWCal pubevents servlets. 

   @author Greg Barnes
   @version 1.0
*/
public abstract class PubEventsServlet extends UWServlet 
{
  /** The user making the request */
  private String uid;

  /** Constructor */
  public PubEventsServlet() {
    super(new CalLogFactory());
  }

  /** @return The user making the request */
  protected String getUid()
  {
    return this.uid;
  }

  /** @return The printable name of the servlet */
  // can be overridden if you want something that looks nicer
  protected String getName() {
    return this.getClass().getName();
  }

  /**
    Initialize the servlet
    @param conf Configuration information
    @exception ServletException If there's a problem
   */
  public void init(ServletConfig conf) throws ServletException
  {
    super.init(conf);
    createLog().println("Starting servlet " + getName());
  }

  /**
    Service an HTTP GET request
    @param req HTTP request object
    @param resp HTTP response object
    @exception ServletException If there's a problem with the servlet
    @exception IOException If there's a problem writing the output
   */
  public void doGet(HttpServletRequest req, HttpServletResponse resp)
      throws ServletException, IOException
  {
    doPost(req, resp);
  }

  /** @return The URL of the login page */
  protected abstract String getStartURL();

  /** @return the name of the 'not authorized' file */
  protected abstract String getNotAuthorizedFile();

  /** 
    Redirect the user back to the login page
    @param req HTTP request object
    @param resp HTTP response object
    @exception IOException If there's a problem writing the output
   */
  private void backToStart(HttpServletRequest req, HttpServletResponse resp) 
      throws IOException
  {
    sendRedirect(req, resp, getStartURL(), false);
  }

  /** Perform authorization
    @param req HTTP request object
    @param resp HTTP response object
    @exception Exception If there's a problem 
   */
  protected abstract boolean authorized(String uid)
      throws Exception;

  /** Perform the main action of the servlet
    @param req HTTP request object
    @param resp HTTP response object
    @exception Exception If there's a problem 
   */
  protected abstract void performAction(HttpServletRequest req, 
                                        HttpServletResponse resp)
      throws Exception;

  /**
    Get the user making the request
    @param req HTTP request object
    @exception InvalidSession If there's a problem with the session
    @exception Exception If there's a problem
   */
  protected abstract String getUser(HttpServletRequest req)
      throws InvalidSession, Exception;

  /**
    Service an HTTP POST request
    @param req HTTP request object
    @param resp HTTP response object
    @exception ServletException If there's a problem invoking the servlet
    @exception IOException If there's a problem writing the output
   */
  public void doPost(HttpServletRequest req, HttpServletResponse resp)
      throws ServletException, IOException
  {
    ServletSupport ss = new ServletSupport(getName(), new PEMessageMaker(), 
                                           new CalLogFactory(), true);
    try {
      uid = getUser(req);

      if (uid == null) {
        backToStart(req, resp);
      }

      resp.setContentType("text/html");
      if (!authorized(uid)) {
        PrintWriter out = resp.getWriter();
  
        try {
          new PrintWebPage(getNotAuthorizedFile(), out);
          out.close();
          return;
        } finally {
          out.close();
        }
      }

      performAction(req, resp);
    } catch (InvalidSession e){
      new CalLogFactory().create(uid).println(e);
      new PEMessageMaker().sessionErrorMessage(resp);
    } catch (Throwable e) {
      ss.handleException(e, resp);
    } finally {
      ss.done();
    }
  }

  /** 
    @return Whether a user is authorized to log in with administrator previ.
    @param uid User to check 
    @exception SQLException If there's a problem checking the database
   */
  protected boolean adminAuthorized(String uid) throws SQLException
  {
    int userid;
    SqlStatements sqlstmts = new SqlStatements();

    try {
      ResultSet rs = sqlstmts.selectUsers(uid);
      if (!rs.next()) {
         return false;
      } else {
         userid = rs.getInt(Fields.UUSER_ID.getName());
         rs.close();
         rs = sqlstmts.selectAuth(userid);
      return rs.next() && rs.getString(Fields.AUSERTYPE.getName()).equals("A");
      }
    } finally {
      sqlstmts.close();
    }
  }

  /** 
    @return Whether a user is authorized to log in
    @param uid User to check 
    @exception SQLException If there's a problem checking the database
   */
  protected boolean pubAuthorized(String uid) throws SQLException
  {
    int userid;
    SqlStatements sqlstmts = new SqlStatements ();

    try {
      ResultSet rs = sqlstmts.selectUsers(uid);
      if (!rs.next()) {
         return false;
      } else {
         userid = rs.getInt(Fields.UUSER_ID.getName());
         rs.close();
         rs = sqlstmts.selectAuth(userid);
         return rs.next();
      }
    } finally {
      sqlstmts.close();
    }
  }
}
