// java

package uwcal.out;

import java.util.Iterator;

import org.apache.ecs.Element;
import org.apache.ecs.ElementContainer;
import org.apache.ecs.Entities;
import org.apache.ecs.html.BR;
import org.apache.ecs.html.TD;
import org.apache.ecs.html.TR;

import edu.washington.cac.calendar.CalLogFactory;
import edu.washington.cac.calendar.MyCalendar;
import edu.washington.cac.calendar.data.Event;
import edu.washington.cac.calendar.data.EventsI;
import edu.washington.cac.calendar.data.InstanceEvent;
import edu.washington.cac.calendar.data.Keyword;
import edu.washington.cac.calendar.data.Location;
import edu.washington.cac.calendar.data.MasterRecurrence;
import edu.washington.cac.calendar.data.NoSuchItemException;
import edu.washington.cac.calendar.data.Recurrence;
import edu.washington.cac.calendar.data.User;
import edu.washington.cac.calendar.db.DBResources;
import edu.washington.cac.myuw.util.Coder;

import uwcal.out.forms.CommonFormElements;

/**
  An object that helps print an event for an EventView
  @author Greg Barnes
  @version 1.1
 */
public class FullEventPrinter extends EventPrinter
{
  /** The user's events cache */
  private EventsI es;

  /** Create a new instance
    @param user User we're creating the instance for
    @param today Current day
    @param es EventsI cache for the user
    @param view View this printer is a part of
   */
  protected FullEventPrinter(User user, 
                             MyCalendar today, 
                             EventsI es, 
                             Timeview view) 
  {
    super(user, today, view);
    this.es = es;
  }

  protected String linkStyle() 
  {
    return Style.LINK_ITEM;
  }

  protected String spanStyle() 
  {
    return Style.DAY_ITEM;
  }

  protected String noTimeRange() 
  {
    return "";
  }

  /**
    Is this event being printed in the context of a single day
    @return Is this event being printed in the context of a single day?
   */
  protected boolean singleDayView()
  {
    return getFirstDay().isSameDay(getLastDay());
  }

  /**
    Assuming we're printing something for the date/time,
    should we see a hyphen (between the start and end dates)?

    @param e Event we're printing
    @return Should we see a hyphen (between the start and end dates)?
   */
  protected boolean shouldPrintHyphen(Event e)
  {
    if (singleDayView()) {
      return super.shouldPrintHyphen(e);
    } else {
      return e.endCalendar() != null && 
             (e.endCalendar().hasTime() ||
              !e.endCalendar().isSameDay(e.startCalendar()));
    }
  }

  /**
    Get the date/time info for the event
    @param e The event
    @return the date/time info for the event
   */
  protected Element timeRange(Event e)
  {
    if (singleDayView()) {
      return super.timeRange(e);
    } else if (hasNoTime(e)) {
      return new ElementContainer().addElement(noTimeRange());
    } else {
      ElementContainer ec = new ElementContainer().
          addElement(dateTime(e.startCalendar(), true));

      if (shouldPrintHyphen(e)) {
        ec.addElement(" - ").
           addElement(dateTime(e.endCalendar(), true));
      }

      return Util.spanItem(ec, timeStyle());
    }
  }

  /**
    Does an event have no date/time information to print?
    @param e The event
    @return Does an event have no date/time information to print?
   */
  protected boolean hasNoTime(Event e) 
  {
    if (singleDayView()) {
      return super.hasNoTime(e);
    } else {
      return e.startCalendar() == null;
    }
  }

  private Element longDesc(Event e) 
  {
    if (e.getLongdesc() == null) {
      return new ElementContainer(); 
    } else {
      return new TR().
          addElement(new TD().addElement(Coder.HTMLencode(e.getLongdesc())));
    }
  }

  private Element cost(Event e) 
  {
    if (e.getCost() == null) {
      return new ElementContainer(); 
    } else {
      return new TR().addElement(new TD().
          addElement(Util.description(DisplayText.COST)).
          addElement(Entities.NBSP).addElement(Coder.HTMLencode(e.getCost())));
    }
  }

  /* Note that for locations, sponsors, and keywords, we only
     print public/private info if the event is private (last arg to 
     constructors of LocationPrinter, SponsorPrinter and KeywordPrinter).  
     If the event is public, these must all be public.  And any 
     responsibility that might rest with the creator of these objects 
     devolves to the creator of the event who chose them.
   */
  /**
    @return HTML for the location of an event
    @param e The event
   */
  protected Element location(Event e) 
  {
    return new TR().
           addElement(new TD().
               addElement(Util.description(DisplayText.LOCATIONF)).
               addElement(Entities.NBSP).
               addElement(new LocationPrinter(today, this.user, !e.isPublic()).
                          format(e.getLocation(this.es), this.view)));
  }

  /**
    @return HTML for the sponsor of an event
    @param e The event
   */
  private Element sponsor(Event e) {
    return new TR().
           addElement(new TD().
               addElement(Util.description(DisplayText.SPONSORF)).
               addElement(Entities.NBSP).
               addElement(new SponsorPrinter(today, this.user, !e.isPublic()).
                          format(e.getSponsor(this.es), this.view)));
  }

  /** 
    Should this keyword be printed?
    @param k Keyword in question
    @return Should this keyword be printed?
   */
  private boolean shouldPrint(Keyword k)
  {
    return !k.getCreator().equals(DBResources.notToBeLoadedByDefaultCreator());
  }

  /**
    Get HTML for the keywords of an event
    @param e The event
    @return HTML for the keywords of an event
   */
  private Element keywords(Event e) {
    TD cell = new TD().
              addElement(Util.description(DisplayText.KEYWORDSF));
    Iterator en = e.getKeywords();
    boolean missingKeywords = false; 
      // if true, keyword Vector needs cleanup
    boolean printedAKeyword = false;

    if (en.hasNext()) {
      do {
        int id = ((Integer) en.next()).intValue();

        try {
          Keyword k = es.getKeywords().getKeyword(id);

          if (shouldPrint(k)) {
            printedAKeyword = true;
            cell.addElement(new BR()).addElement("\n").
                 addElement(new KeywordPrinter(today, this.user, !e.isPublic()).
                            format(k, this.view));
          }
        } catch (NoSuchItemException e1) {
          new CalLogFactory().create().println("Event " + e.getId() + 
                                               " missing keyword " + id);
          missingKeywords = true;
        }
      } while (en.hasNext());

      if (missingKeywords) {
        e.deleteMissingKeywords(es.getKeywords());
      }
    } 

    if (!printedAKeyword) {
      cell.addElement(Entities.NBSP).addElement(DisplayText.NO_VALUE);
    }

    return new TR().addElement(cell);
  }

  /**
    Get the master event for a recurring event
    @param e Event that is an instance of a recurrence
    @return The master event for the recurrence
   */
  private Event master(Event e)
  {
    if (e.getRecurrence().isMaster()) {
      return e;
    } else {
      return ((InstanceEvent) e).getMaster();
    }
  }

  /**
    Get HTML for a single recurrence instance
    @param e Event to print
    @return HTML for a single recurrence instance
   */
  private Element printInstance(Event e)
  {
    return Util.linkItem(Eventview.url(e, e.startCalendar(), this.view),
                         timeRange(e));
  }

  /**
    Get HTML for a list of recurrence instances, not including the current event

    @param e The event we're printing the list for
    @param recur A Master recurrence
    @return HTML for a list of recurrence instances
   */
  private Element recurList(Event e, MasterRecurrence recur)
  {
    if (recur.knownInstances() <= 1) {
      return new ElementContainer();
    }

    ElementContainer list = new ElementContainer().
                            addElement(".  " + DisplayText.RECURLIST_START);

    Iterator i = recur.getInstances();

    while (i.hasNext()) {
      Event e2 = (Event) i.next();

      if (e2 != e) {
        list.addElement(new BR()).addElement(printInstance(e2));
      }
    }

    return list;
  }

  /**
    Get HTML for any recurrence information about the event
    @param e The event
    @return HTML for any recurrence information about the event
   */
  private Element recurInfo(Event e)
  {
    if (!e.isRecurring()) {
      return new ElementContainer();
    } else {
      return new TR().addElement(new TD().
          addElement(DisplayText.RECURRING_EVENT).
          addElement(recurList(e, 
              (MasterRecurrence) (master(e).getRecurrence()))));
    }
  }

  /**
    Get HTML to allow the user to add the event to their calendar,
    if that is appropriate
    @param e event in question
    @return HTML for the line
   */
  protected Element addEventRefLine(Event e)
  {
    if (user.isGuest()) {
      return new TR().addElement(new TD().
          addElement(new AddEventRef().addLine(e)));
    } else {
      return new ElementContainer();
    }
  }

  /** 
    Get HTML for the page for an event
    @return HTML for the page 
    @param e Event to print
    @param es user's cache of events
   */
  protected Element format(Event e, EventsI es) 
  {
    return new ElementContainer().
           addElement(CommonFormElements.headerRow(
               DisplayText.EVENTVIEW_HEADER, Style.FORM_TITLE)).
           addElement(new TR().
                      addElement(new TD().
                                 addElement(timeRange(e)).addElement("\n ").
                                 addElement(hasNoTime(e) ? "" : Entities.NBSP).
                                 addElement(eventDesc(e)).addElement("\n"))).
           addElement(longDesc(e)).addElement("\n").
           addElement(new TR().
                      addElement(new TD().
                          addElement(new PublicPrivatePrinter(today, this.user).
                          format(new PPEventPrinter(e, this.view))))).
           addElement("\n").
           addElement(cost(e)).addElement("\n").
           addElement(location(e)).addElement("\n").
           addElement(sponsor(e)).addElement("\n").
           addElement(keywords(e)).addElement("\n").
           addElement(recurInfo(e)).addElement("\n").
           addElement(addEventRefLine(e)).addElement("\n");
  }
}
