// java

package uwcal.out;

import edu.washington.cac.calendar.MyCalendar;
import edu.washington.cac.calendar.data.Cacheable;
import edu.washington.cac.calendar.data.PPEntity;
import edu.washington.cac.calendar.data.PPEntityUtil;
import edu.washington.cac.calendar.data.User;
import edu.washington.cac.calendar.db.DBResources;

/** Something that helps print a PPEntity 
   @author Greg Barnes
   @version 1.2
 */
public abstract class PPEPrinter 
{
  /** The entity to print */
  private PPEntity ppe;

  /** Page this printer is a part of */
  protected Page page;

  /** Create a new instance
    @param ppe Entity to print
    @param page Page this printer is a part of 
   */
  PPEPrinter(PPEntity ppe, Page page)
  {
    this.ppe = ppe;
    this.page = page;
  }

  /** 
    Can this entity be edited by a user
    @param user The user in question
    @return Can this entity be edited by a user
   */
  boolean canBeEditedBy(User user) 
  {
    return PPEntityUtil.canEdit(user, this.ppe);
  }
        
  /** @return the creator of the entity */
  User getCreator() 
  {
    return this.ppe.getCreator();
  }

  /** 
    is the entity public? 
    @return is the entity public? 
   */
  boolean isPublic() 
  {
    return this.ppe.isPublic();
  }

  /** 
    Should the creator of this entity be printed?
    @return Should the creator of this entity be printed?
   */
  boolean shouldPrintCreator()
  {
    return isPublic() && 
           !getCreator().toString().
               startsWith(DBResources.specialCreator().toString());
  }

  abstract String deleteUrl(MyCalendar today, int sid);
  abstract String editUrl(MyCalendar today, int sid);
  abstract String ppeName();

  /* The following 2 functions should be overridden for entities that can
     be removed */

  /** 
    Can the entity be removed (c.f. The remove function)? 
    @return Can the entity be removed (c.f. The remove function)? 
   */
  boolean canBeRemoved()
  {
    return false;               // default is no
  }

  /** 
    Get The URL to remove the entity 
    @param today The current day
    @param sid The session identifier
    @return The URL to remove the entity 
   */
  String removeUrl(MyCalendar today, int sid) 
  {
    System.err.println("this removeUrl should not be called");
    return "";
  }

  /* The following 2 functions should be overridden for entities that can
     be removed, and can have multiple instances
   */
  /**
    Is the entity referenced one instance of many?
    @return Is the entity referenced one instance of many?
   */
  boolean hasMultipleInstances()
  {
    return false;
  }

  /** 
    Get The URL to remove all instances of the entity 
    @param today The current day
    @param sid The session identifier
    @return The URL to remove all instances of the entity 
   */
  String removeAllUrl(MyCalendar today, int sid) 
  {
    System.err.println("this removeAllUrl should not be called");
    return "";
  }

  /* The following function should be overridden for entities that are
     official */

  /** 
    is the entity 'official'? 
    @return is the entity 'official'? 
   */
  boolean isOfficial() 
  {
    return false;            
  }
}
